package codebreaking

object MasterMind:
  enum Color:
    case Red, Purple, Brown, Blue, White, Pink, Green, Carmine, Fuchsia

  type Word = List[Color]
  case class Score(redKeys: Int, whiteKeys: Int)
  val allColors: List[Color] = Color.values.toList

  class Game(wordLength: Int) extends Puzzle[Word, Score]:

    def score(guess: Word, secret: Word): Score =
      require(guess.length == wordLength, f"Guess must have length $wordLength.")
      require(secret.length == wordLength, f"Secret must have length $wordLength.")
      val exact = guess.zip(secret).count(_ == _)
      val all = allColors.map(c => math.min(guess.count(_ == c), secret.count(_ == c))).sum
      Score(exact, all - exact)

    lazy val allSecrets = iterWords(allColors.toSet, wordLength)

    lazy val validGuesses = allSecrets

    def occSet(word: Word) =
      word.groupBy(identity).view.mapValues(_.length)
        .toMap.withDefaultValue(0)
    def score2(guess: Word, secret: Word): Score =
      require(guess.length == secret.length)
      val exact = guess.zip(secret).count(_ == _)
      val (guessSet, secretSet) = (occSet(guess), occSet(secret))
      val all = guessSet.keySet.union(secretSet.keySet)
        .map(c => math.min(guessSet(c), secretSet(c))).sum
      Score(exact, all - exact)

    def score3(guess: Word, secret: Word): Score =
      require(guess.length == secret.length)
      val exact = guess.zip(secret).count(_ == _)
      val all = guess.foldLeft((0, secret)) {
        case ((matches, leftover), c) =>
          val (before, after) = leftover.span(_ != c)
          if after.isEmpty then (matches, leftover)
          else (matches + 1, before ++ after.tail)
      }._1
      Score(exact, all - exact)

    def guessQuality(score: Score): Int =
      2 * score.redKeys + score.whiteKeys
