package codebreaking

object Wordle:
  type Word = String

  enum Result:
    case Grey, Yellow, Green
  type Score = List[Result]

  val allLetters = ('a' to 'z').toSet

  class Game(wordLength: Int, wordList: Set[Word]) extends Puzzle[Word, Score]:

    def score(guess: Word, secret: Word): Score =
      require(guess.length == wordLength, f"Guess length must be $wordLength.")
      require(secret.length == wordLength, f"Secret length must be $wordLength.")
      require(guess.distinct == guess, f"Guess $guess cannot have duplicates.")
      require(secret.distinct == secret, f"Secret $secret cannot have duplicates.")
      guess.zip(secret).toList.map { (g, s) =>
        if g == s then Result.Green
        else if secret.contains(g) then Result.Yellow
        else Result.Grey
      }

    lazy val allSecrets = wordList.filter(w => w.length == wordLength && w.distinct == w)

    lazy val validGuesses = allSecrets

    def scoreWithDuplicates(guess: Word, secret: Word): Score =
      require(guess.length == secret.length, "Guess must have the same length as secret.")
      guess.zip(secret).foldLeft[(Set[Char], Score)]((secret.toSet, List())) {
        case ((left, graded), (g, s)) =>
          (
            left - g,
            (if g == s then Result.Green
             else if left.contains(g) then Result.Yellow
             else Result.Grey) +: graded
          )
      }._2.reverse

    def guessQuality(score: Score): Int =
      score.map {
        case Result.Grey   => 0
        case Result.Yellow => 1
        case Result.Green  => 2
      }.sum
